import discord
import discord.ext.commands
from discord import app_commands

from .interaction_history import InteractionHistory

__all__ = ('RacketBot',)


class RacketBot(discord.Client):

    def __init__(self,
                 *,
                 intents: discord.Intents,
                 guild_ids: list[int] = None,
                 force_command_sync: bool = False):
        super().__init__(intents=intents)
        self.tree = app_commands.CommandTree(self)
        self.guild_ids = guild_ids
        self.force_command_sync = force_command_sync
        self.extra_events = {}
        self.interaction_history = InteractionHistory()

    async def setup_hook(self):
        if self.force_command_sync:
            await self.sync_command_definitions()

    async def sync_command_definitions(self):
        if self.guild_ids is not None:
            for guild_id in self.guild_ids:
                self.tree.copy_global_to(guild=discord.Object(id=guild_id))
                await self.tree.sync(guild=discord.Object(id=guild_id))
        else:
            pass

    # Lightweight re-implementation of discord.ext.commands.Bot.add_cog
    # We only care about the app commands inside the cog, not any other type of
    # command/command group.
    #
    # We don't want to inherit from discord.ext.commands.Bot because that type
    # of bot requires a prefix in addition to supporting slash commands which is
    # more than we want.
    # We have to re-implement because we inherit from discord.Client instead.
    #
    def add_cog(self, cog: discord.ext.commands.Cog):
        for cmd in cog.get_app_commands():
            self.tree.add_command(cmd)
        for name, listener in cog.get_listeners():
            if name not in self.extra_events:
                self.extra_events[name] = []
            self.extra_events[name].append(listener)

    # Client does not support extra listeners very well. This is a copy of
    # discord.ext.commands.Bot.dispatch which mostly just calls super() but will
    # also dispatch to our extra_events.
    def dispatch(self, event_name: str, *args, **kwargs):
        super().dispatch(event_name, *args, **kwargs)
        ev = 'on_' + event_name
        for event in self.extra_events.get(ev, []):
            self._schedule_event(event, ev, *args, **kwargs)

    async def on_interaction(self, interaction: discord.Interaction):
        self.interaction_history.record_usage(interaction)


import os
import logging

import discord

__all__ = ('setup_logging',)

# TODO: Filter the "ffmpeg process %s succefully terminated" messages.
# TODO: Figure out why there are format strings in the text log, and remove them.


def check_or_create_logs_dir():
    if not os.path.exists('logs/'):
        os.makedirs('logs/')


def setup_cli_logs():
    # Default logging to the CLI with color.
    cli_handler = logging.StreamHandler()
    cli_handler.setLevel(logging.INFO)
    # Need to set level=logging.DEBUG because that's the logger setting.
    # The cli_handler will filter out those LogRecords below logging.INFO
    discord.utils.setup_logging(handler=cli_handler, level=logging.DEBUG)


def setup_file_logs():
    # Add some file logging at a higher verbosity.
    file_handler = logging.FileHandler(filename='logs/discord.log',
                                       encoding='utf-8',
                                       mode='a')
    file_handler.setLevel(logging.INFO)
    discord.utils.setup_logging(handler=file_handler, level=logging.DEBUG)


def setup_logging():
    check_or_create_logs_dir()
    setup_cli_logs()
    setup_file_logs()


"""Utility for running a cog. See run_cog docstring for more details."""
from collections.abc import Sequence
import logging
from typing import Type

from racket import setup_logging, RacketBot

import discord
import discord.ext.commands


__all__ = ('run_cog',)


def run_cog(cog_class: Type[discord.ext.commands.Cog],
            *,
            token: str,
            guilds: Sequence[int] | None = None) -> None:
    """Runner function for your cog.

    Most bots implement a single Cog. This utility abstracts away all the logging,
    data, and bot construction, just running your cog for you.

    Your cog should inherit from discord.ext.commands.Cog, and it should take a
    single argument `racket.RacketBot`. racket.run_cog(...) will do the instatiation
    for you. This is to ensure that all logging is setup before the cog is
    initialized.

    ```
    import racket

    class MyCog(discord.ext.commands.Cog):
        def __init__(self, bot: racket.RacketBot):
            ...

    racket.run_cog(MyCog)
    ```
    """
    # Most importantly, do this as early as possible.
    setup_logging()
    log = logging.getLogger(__name__)

    intents = discord.Intents.default()
    intents.typing = False  # pylint: disable=assigning-non-slot
    intents.presences = False  # pylint: disable=assigning-non-slot
    intents.message_content = False  # pylint: disable=assigning-non-slot

    log.info('Creating bot object.')
    bot = RacketBot(intents=intents, guild_ids=guilds)

    log.info('Adding custom commands.')
    bot.add_cog(cog_class(bot))

    log.info('Starting the bot.')
    # Set the log_handler to avoid double logging settup.
    bot.run(token, log_handler=None)



from collections.abc import Sequence
import datetime
import json
import os
import sqlite3

import discord

_DB_DIR = 'data/'
_DB_NAME = 'interaction_history.db'
_TABLE_NAME = 'interaction_history'

__all__ = ('InteractionHistory',)


class InteractionHistory:
    """Stores a history of ApplicationInteractions."""

    def __init__(self):
        if not os.path.exists(_DB_DIR):
            os.makedirs(_DB_DIR)
        self._con = sqlite3.connect(_DB_DIR + _DB_NAME)
        self._create_table_if_missing()

    def _create_table_if_missing(self):
        cur = self._con.cursor()
        res = cur.execute('SELECT name FROM sqlite_master').fetchone()
        if res is not None and _TABLE_NAME in res:
            return
        cur.execute(
            f'CREATE TABLE {_TABLE_NAME}(datetime, user_id, guild_id, channel_id, message_id, type, raw_data)'
        )
        self._con.commit()
        cur.close()

    def record_usage(self, interaction: discord.Interaction) -> None:
        """Records someone triggering and interaction."""
        now = datetime.datetime.now(datetime.timezone.utc).isoformat()
        cur = self._con.cursor()

        message_id = 'None'
        if interaction.message is not None:
            message_id = str(interaction.message.id)

        cur.execute(f'INSERT INTO {_TABLE_NAME} VALUES(?, ?, ?, ?, ?, ?, ?)',
                    (now, interaction.user.id, str(interaction.guild_id),
                     str(interaction.channel_id), message_id,
                     interaction.type.name, json.dumps(interaction.data)))
        self._con.commit()
        cur.close()

    def fetch_all_history(
            self) -> Sequence[tuple[datetime.datetime, int, str, dict]]:
        """Return all the history.
        
        Each row is in this format:
        (utc_datetime, user_id, type, raw_data)
        """
        cur = self._con.cursor()

        def none_or_int(x: None | str) -> None | int:
            if x.lower() == 'none':
                return None
            return int(x)

        res = [{
            'datetime': datetime.datetime.fromisoformat(row[0]),
            'user_id': row[1],
            'guild_id': none_or_int(row[2]),
            'channel_id': none_or_int(row[3]),
            'message_id': none_or_int(row[4]),
            'type': row[5],
            'raw_data': json.loads(row[6]),
        } for row in cur.execute(
            f'SELECT datetime, user_id, guild_id, channel_id, message_id, type, raw_data FROM {_TABLE_NAME}'
        )]
        res.sort(key=lambda x: x['datetime'], reverse=True)
        return res
